# Tool that invokes a script in a project local bin directory.
#
# The scripts are called with the project path as the first argument.
import logging

import menhir.tool as mtool

log = logging.getLogger(__name__)


class Script():
    pass


def tool():
    return Script()


@mtool.name.register(Script)
def name(arg):
    return "script"


@mtool.is_using_tool.register(Script)
def is_using_tool(tool, path):
    from os.path import exists, join
    return exists(join(path, 'bin'))


@mtool.dependencies.register(Script)
def dependencies(tool, path):
    return []


@mtool.execute_build_phase.register(Script)
def execute_build_phase(tool, path, phase_name):
    """Execute a build phase."""
    import os
    import os.path
    import subprocess

    script_file = os.path.join(path, 'bin', phase_name)
    if os.access(script_file, os.X_OK):
        result = subprocess.run([script_file, path])
        if result.returncode:
            return {
                'status': 'fail',
            }
        return {
            'status': 'ok',
        }
    else:
        log.debug('No script for build phase %s', phase_name)
        return {
            'status': 'nothing_to_do'
        }



# pytest based tool for menhir
import menhir.tool as mtool


class Pytest():
    pass


def tool():
    return Pytest()


@mtool.name.register(Pytest)
def name(arg):
    return "pytest"


@mtool.is_using_tool.register(Pytest)
def is_using_tool(tool, path):
    from os.path import exists, join
    path = join(path, 'setup.cfg')
    if exists(path):
        with open(path, "r") as file:
            data = file.readlines()
            return 'pytest' in data


@mtool.dependencies.register(Pytest)
def dependencies(tool, path):
    return []


@mtool.execute_build_phase.register(Pytest)
def execute_build_phase(tool, path, phase_name):
    """Execute a build phase."""
    import os
    import subprocess
    if phase_name == 'test':
        owd = os.getcwd()
        try:
            os.chdir(path)
            res = subprocess.run(['pytest'])
            if res.returncode:
                return {
                    'status': 'fail'
                }
            return {
                'status': 'ok',
            }
        finally:
            os.chdir(owd)
    else:
        return {
            'status': 'nothing_to_do'
        }



import logging

log = logging.getLogger(__name__)


def main():
    args = parser().parse_args()
    logging.basicConfig(level=args.log_level)
    if args.phase:
        exec_phase(args.phase, args.verbose, args.commit, args.circleci)


def parser():
    import argparse
    parser = argparse.ArgumentParser(description='Menhir monorepo build tool.')
    parser.add_argument("--commit")
    parser.add_argument("-v", "--verbose", default=False,
                        action='store_const', const=True)
    parser.add_argument("--circleci", default=False,
                        action='store_const', const=True)
    parser.add_argument("--log_level", default='WARNING')
    subparsers = parser.add_subparsers(help='sub-commands')
    parser_exec = subparsers.add_parser('exec', help='execute a build phase')
    parser_exec.add_argument('phase')
    return parser


def exec_phase(phase, verbose, commit, circleci):
    import sys
    from menhir import build_order, execute_phase, root_config
    config = root_config()

    project_dirs = config['project_dirs']
    if verbose:
        print('Project dirs:')
        for p in project_dirs:
            print('    ', p)

    from_commit, to_commit = commit_range(
        commit, circleci, verbose
    )
    if verbose:
        print("Commit range: %s..%s" % (from_commit, to_commit))

    g = build_order(
        config['dependencies'],
        config['project_dirs'],
        config['root_projects'],
        from_commit,
        to_commit,
    )

    if verbose:
        print('Build phase:', phase)
    res = execute_phase(g, config['tool_impls'], phase)

    if res['status'] == 'fail':
        print('Build phase "%s" failed' % phase, file=sys.stderr)
        sys.exit(1)
    else:
        sys.exit(0)


def commit_range(commit_arg, circleci, verbose):
    import os
    from menhir.gitutils import head_commit, repo
    if commit_arg:
        return commits_from_range(commit_arg)
    if circleci:
        compare_url = os.getenv('CIRCLE_COMPARE_URL')
        log.debug('Deriving commit range from %s', compare_url)
        if compare_url:
            commit_range = compare_url.rsplit('/', maxsplit=1)[1]
            return commits_from_range(commit_range)
        sha = os.getenv('CIRCLE_SHA1')
        return sha+"^", sha
    commit = head_commit(repo())
    return str(commit)+"^", str(commit)


def commits_from_range(commit_range):
    log.debug('commits_from_range: %s', commit_range)
    commits = commit_range.split('..')
    if len(commits) == 2:
        from_commit = commits[0].replace('.', '')
        to_commit = commits[1].replace('.', '')
        log.debug('commits_from_range: from %s to %s', from_commit, to_commit)
        return from_commit, to_commit
    log.debug('commits_from_range: single commit %s', commit_range)
    return commit_range+"^", commit_range



# Setup.py based tool for menhir
import logging
import re
import sys

import menhir.tool as mtool

DEP_PATTERN = re.compile('file:(.*)#egg=.*')

log = logging.getLogger(__name__)


class SetupPy():
    pass


def tool():
    return SetupPy()


@mtool.name.register(SetupPy)
def name(arg):
    return "setup_py"


@mtool.is_using_tool.register(SetupPy)
def is_using_tool(tool, path):
    from os.path import exists, join
    return exists(join(path, 'setup.py'))


@mtool.dependencies.register(SetupPy)
def dependencies(tool, path):
    from os.path import join, normpath
    log.debug('setup.py dependencies in %s', path)
    try:
        requirements_path = join(path, 'requirements.txt')
        requirements, dependency_links = read_requirements(requirements_path)
        return [
            normpath(join(path, DEP_PATTERN.match(l).group(1)))
            for l in dependency_links
            if l.startswith('file:')
        ]
    except:
        print('Failed to infer dependencies in for setup_py %s' % path,
              file=sys.stderr)
        raise


def read_requirements(path):
    with open(path, 'r') as f:
        return process_requirements(f)


def process_requirements(f):
    requirements = []
    dependency_links = []
    for line in f.read().split():
        if '#egg=' in line:
            package_name = line.rsplit('#egg=', 1).pop()
            requirements.append(package_name)
            dependency_links.append(line)
        else:
            requirements.append(line)
    return requirements, dependency_links


@mtool.execute_build_phase.register(SetupPy)
def execute_build_phase(tool, path, phase_name):
    """Execute a build phase."""
    return {
        'status': 'nothing_to_do'
    }



# GIT utilities

from git import Repo


def repo(base_dir=None):
    """Return a repository object."""
    import os
    base_dir = base_dir or os.getcwd()
    return Repo(base_dir)


def commit(repo, commitish):
    """Return a commit object."""
    return repo.commit(commitish)


def head_commit(repo):
    """Return a commit object."""
    return repo.head.commit


def diff(start, end):
    """Diff for the commit range."""
    return diff_paths(start.diff(end))


def diff_paths(diff):
    rpaths = [p.a_path for p in diff.iter_change_type('R')]
    rpaths.extend([p.b_path for p in diff.iter_change_type('R')])

    diffs = {
        'added': [p.b_path for p in diff.iter_change_type('A')],
        'deleted': [p.a_path for p in diff.iter_change_type('D')],
        'renamed': rpaths,
        'modified': [p.b_path for p in diff.iter_change_type('M')],
    }
    all_paths = []
    for i in ['added', 'deleted', 'renamed', 'modified']:
        all_paths.extend(diffs[i])
    diffs['all'] = all_paths
    return diffs



# Build tools that menhir knows how to use.  This is a mapping from
# name to namespace used to implement the support.
import logging

from menhir.config import root_config

log = logging.getLogger(__name__)

tools = {
    'setup.py': 'menhir.tools.setup_py'
}


def execute_phase(build_order, tool_impls, phase_name):
    import menhir.tool as mtool
    for projects in build_order:
        for path in reversed(projects):
            for tool in tool_impls:
                res = mtool.execute_build_phase(tool, path, phase_name)
                log.debug('Execute tool %s phase %s in %s returns %s',
                          mtool.name(tool), phase_name, path, res)
                if res['status'] == 'fail':
                    log.debug('Tool %s phase %s in %s failed %s',
                              mtool.name(tool), phase_name, path, res)
                    res['tool'] = mtool.name(tool)
                    res['phase'] = phase_name
                    res['path'] = path
                    return res
    return {
        'status': 'ok'
    }


def build_order(dependencies, project_dirs, root_projects, start, end):
    from menhir.graph import dfs_paths
    g = changed_graph(dependencies, project_dirs, root_projects, start, end)
    roots = [node for node in root_projects if node in g]
    paths = [node for node in dfs_paths(g, roots)]
    return filtered_build_order(paths)


def filtered_build_order(build_order):
    x = [set()]
    x.extend(build_order)
    rs = reductions(set.union, map(set, x), set())
    return list(map(
        lambda x, y: list(filter(lambda z: z not in y, x)),
        build_order,
        rs
    ))


def reductions(f, lst, init):
    prev = init
    for i in lst:
        prev = f(prev, i)
        yield prev


def changed_graph(dependencies, project_dirs, root_projects, start, end):
    from menhir.graph import build_graph
    projs = changed_projects(project_dirs, start, end)
    return build_graph(dependencies, projs, root_projects)


def changed_projects(roots, start, end):
    """Find the set of projects that contain changed files."""
    files = changed_files(start, end)['all']
    return projects_with_files(roots, files)


def projects_with_files(roots, files):
    roots = set(roots)
    changed_roots = set()
    log.debug('roots: %s', roots)
    log.debug('files: %s', files)
    for path in files:
        if len(roots):
            for root in roots.copy():
                if path.startswith(root):
                    changed_roots.add(root)
                    roots.remove(root)
        else:
            break
    return changed_roots


def changed_files(start, end):
    from menhir import gitutils
    repo = gitutils.repo()
    start_commit = gitutils.commit(repo, start)
    end_commit = gitutils.commit(repo, end)
    changed_files = gitutils.diff(start_commit, end_commit)
    return changed_files


__all__ = [
    'root_config',
    'changed_files',
]



# pyenv based tool for menhir
import logging

import menhir.tool as mtool

log = logging.getLogger(__name__)


class Pyenv():
    pass


def tool():
    return Pyenv()


@mtool.name.register(Pyenv)
def name(arg):
    return "pyenv"


@mtool.is_using_tool.register(Pyenv)
def is_using_tool(tool, path):
    from os.path import exists, join
    path = join(path, 'requirements.txt')
    return exists(path)


@mtool.dependencies.register(Pyenv)
def dependencies(tool, path):
    return []


def run(cmd, *args, **kwargs):
    import subprocess
    res = subprocess.run(cmd, *args, *kwargs)
    if res.returncode:
        return False, {
            'status': 'fail'
        }
    return True, {
        'status': 'ok',
    }


@mtool.execute_build_phase.register(Pyenv)
def execute_build_phase(tool, path, phase_name):
    """Execute a build phase."""
    from os import chmod, remove
    from os.path import basename
    import stat
    import subprocess
    import tempfile
    import pkg_resources

    print('exec', path, phase_name)
    if phase_name == 'requirements':
        log.debug('pyenv requirements in %s', path)
        resource_package = "menhir"
        resource_path = "/tools/pyenv/requirements.sh"
        script = pkg_resources.resource_string(resource_package, resource_path)
        fname = None
        try:
            with tempfile.NamedTemporaryFile("xb", delete=False) as f:
                fname = f.name
                f.write(script)
            chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
            res = subprocess.run([f.name, path, basename(path)])
        finally:
            remove(fname)

        if res.returncode:
            return {
                'status': 'fail'
            }
        return {
            'status': 'ok'
        }
    else:
        return {
            'status': 'nothing_to_do'
        }



# Graph functions

# Graphs are modelled as a dict of edges from source node to a set of
# child nodes.


def root_nodes(graph, nodes):
    """Return the set of all nodes that are not dependencies of other nodes."""
    from functools import reduce
    nodes = set(nodes)
    return reduce(set.difference, graph.values(), nodes)


def dfs_paths(graph, root_nodes):
    stack = [(node, [node]) for node in root_nodes]
    while stack:
        (vertex, path) = stack.pop()
        next_vertices = graph.get(vertex, set()) - set(path)
        for next in next_vertices:
            stack.append((next, path + [next]))
        if not next_vertices:
            yield path


def build_graph(graph, changed, root_nodes):
    stack = [(node, [node]) for node in root_nodes]
    res = {}
    while stack:
        (vertex, path) = stack.pop()
        next_vertices = graph[vertex] - set(path)
        for next in next_vertices:
            stack.append((next, path + [next]))
        if not next_vertices:
            i = len(path)
            while i:
                i = i - 1
                if path[i] not in changed:
                    path.pop()
                else:
                    break
            while i > 0:
                from_node = path[i - 1]
                s = graph[from_node] or set()
                s.add(path[i])
                res[from_node] = s
                i = i - 1
            if len(path):
                from_node = path[0]
                res[from_node] = res.get(from_node) or set()
    return res
